// **********************************************************************
//
// Copyright (c) 1998
// Object-Oriented Concepts, Inc.
// Billerica, MA, USA
//
// Copyright (c) 1998
// Object-Oriented Concepts GmbH
// Ettlingen, Germany
//
// All Rights Reserved
//
// **********************************************************************

#ifndef OB_STUB_IMPL_BASE_H
#define OB_STUB_IMPL_BASE_H

namespace PortableServer
{

//
// Forward declaration
//
class ServantBase;

} // End of namespace PortableServer

namespace OB
{

//
// Forward declarations
//
class StubImplBase;
struct CodeConverters;

} // End of namespace OB

namespace OB
{
    
class Client;
typedef Client* Client_ptr;

} // End of namespace OB

void OBDuplicate(OB::Client_ptr);
void OBRelease(OB::Client_ptr);

namespace OB
{
    
//
// Stub implementation changed exception
//
class StubImplChanged
{
};

//
// Location forward exception
//
struct StubImplForward
{
    IOP::IOR ior; // forwarded IOR
    StubImplForward(const IOP::IOR& i) : ior(i) { }
};

//
// The Client class
//
typedef ObjVar< Client > Client_var;
typedef ObjSeq< Client, int > ClientSeq;
typedef SeqVar< ObjSeq< Client, int > > ClientSeq_var;

class Client : public RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    Client(const Client&);
    void operator=(const Client&);

    //
    // The code set converters
    //
    CodeConverters* codeConverters_;

    //
    // If set to true, this client can be reused
    //
    bool reuseClient_;

    //
    // The usage counter
    //
    int usage_;

protected:

    Client();
    virtual ~Client();

public:

    static inline Client_ptr _duplicate(Client_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline Client_ptr _nil()
    { return 0; }

    //
    // Increment usage (not mutex protected)
    //
    void incUsage() { assert(usage_ >= 0); usage_++; }

    //
    // Decrement usage (not mutex protected)
    //
    // Returns true if after the decrement the usage counter is larger
    // than 0, and false otherwise.
    //
    bool decUsage() { assert(usage_ > 0); usage_--; return usage_ > 0; }

    //
    // Get the code set converters
    //
    CodeConverters* getCodeConverters() const;

    //
    // Set the code set converters
    //
    void setCodeConverters(const CodeConverters&);
    
    //
    // Get connection reuse properties
    //
    bool reuseClient() { return reuseClient_; }

    //
    // Set connection reuse properties
    //
    void reuseClient(bool reuse) { reuseClient_ = reuse; }

    //
    // Destroy the client
    //
    virtual void destroy() = 0;

    //
    // Get the OCI Connector
    //
    virtual OCI::Connector_ptr connector() = 0;
    
    //
    // Get the OCI Transport
    //
    virtual OCI::Transport_ptr transport() = 0;

    //
    // Calculate offset
    //
    virtual CORBA::ULong offset(StubImplBase*, const char*) = 0;

    //
    // Send request and receive response
    // With explicit timeout (throws exception on timeout)
    // With optional retry on communication failure
    //
    virtual CORBA::ULong request(StubImplBase*, const char*, Buffer&,
				 bool&, bool&, bool&, CORBA::Long, bool) = 0;

    //
    // Send request oneway
    // With implicit zero timeout (does not throw exception on timeout)
    // With optional retry on communication failure
    //
    virtual void oneway(StubImplBase*, const char*, Buffer&, bool) = 0;

    //
    // Send request deferred
    // With implicit zero timeout (does not throw exception on timeout)
    // With optional retry on communication failure
    //
    virtual void deferred(StubImplBase*, const char*, Buffer&,
			  CORBA::ULong&, bool) = 0;

    //
    // Get repsonse
    // With explicit timeout (throws exception on timeout)
    //
    virtual CORBA::ULong response(CORBA::ULong, Buffer&, bool&, bool&, bool&,
				  CORBA::Long) = 0;
    
    //
    // Poll response
    // With implicit zero timeout (returns zero on timeout)
    //
    virtual CORBA::ULong poll(CORBA::ULong, Buffer&, bool&, bool&, bool&) = 0;
};

} // End of namespace OB

namespace CORBA
{

//
// release and is_nil
//
inline void
release(OB::Client_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::Client_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

namespace OB
{
    
class StubImplBase;
typedef StubImplBase* StubImplBase_ptr;

} // End of namespace OB

void OBDuplicate(OB::StubImplBase_ptr);
void OBRelease(OB::StubImplBase_ptr);

namespace OB
{

//
// The StubImplBase class
//
typedef ObjVar< StubImplBase > StubImplBase_var;

class StubImplBase : public RefCount
{
    //
    // Hide copy-constructor and assignment operator
    //
    StubImplBase(const StubImplBase&);
    void operator=(const StubImplBase&);
    
protected:

    //
    // The client
    //
    Client_var _ob_clt_;

    //
    // The set of policies
    //
    CORBA::PolicyList _ob_policies_;

    //
    // The timeout (in msec)
    //
    CORBA::Long _ob_tout_;

    //
    // Retry flag
    //
    bool _ob_retry_;

    //
    // The IOR
    //
    IOP::IOR _ob_ior_;

    //
    // Key cache
    //
    OctetSeq _ob_key_;

    //
    // For LocalStubImpl
    //

    //
    // The PortableServer::POA
    //
    PortableServer::POA_var _ob_poa_;

    //
    // The PortableServer::ObjectId
    //
    PortableServer::ObjectId _ob_oid_;

    //
    // The objects POA key (required for object-ref -> object-id)
    //
    OctetSeq _ob_poa_key_;

    //
    // This flag is true if the servant has been deactivated
    //
    bool deactivated_;

#ifdef HAVE_JTC
    JTCMutex deactivatedMutex_;
#endif

#ifdef HAVE_VCPLUSPLUS_BUGS
public:
#endif

    //
    // The servant
    //
    PortableServer::ServantBase* _ob_servant_;
    
public:

    StubImplBase();
    virtual ~StubImplBase();

    static inline StubImplBase_ptr _duplicate(StubImplBase_ptr p)
    { if(p) p -> _OB_incRef(); return p; }
    static inline StubImplBase_ptr _nil()
    { return 0; }

    CORBA::InterfaceDef_ptr _get_interface();
    CORBA::Boolean _non_existent();
    CORBA::Boolean _is_a(const char*);
    OCI::TransportInfo_ptr _get_oci_transport_info();
    OCI::ConnectorInfo_ptr _get_oci_connector_info();

    //
    // ORBacus internal functions
    // Application programs must not use these functions directly
    //
    Client_ptr _OB_client();
    void _OB_client(Client_ptr);

    virtual bool _OB_isLocal();

    void _OB_policies(const CORBA::PolicyList&);
    const CORBA::PolicyList& _OB_policies() { return _ob_policies_; }

    CORBA::Policy_ptr _OB_getPolicy(CORBA::PolicyType);

    const IOP::IOR& _OB_ior() { return _ob_ior_; }
    void _OB_ior(const IOP::IOR& ior) { _ob_key_.length(0); _ob_ior_ = ior; }
                                     // ^^^^^^^^^^^^^^^^^^ Invalidate key cache
    const OctetSeq& _OB_key() { return _ob_key_; }
    void _OB_key(const OctetSeq& key) { _ob_key_ = key; }

    const OctetSeq& _OB_POAkey() { return _ob_poa_key_; }
    void _OB_POAkey(const OctetSeq& key)
    {
        _ob_poa_key_ = key;
    }

    CORBA::Long _OB_timeout();
    bool _OB_retry();

    const PortableServer::ObjectId& _OB_oid() { return _ob_oid_; }

    //
    // Mark the stub as deactivated
    //
    void _OB_deactivate();
};

//
// Helper class for setting/clearing down the POA for local
// invocations. When the LocalStub stuff is seperated from
// StubImplBase InvocationHandler should be moved with it.
//
class InvocationHandler
{
    PortableServer::POA_ptr poa_;

public:
    
    InvocationHandler(PortableServer::ServantBase*,
		      PortableServer::POA_ptr,
		      const char*,
		      const PortableServer::ObjectId&,
		      const OctetSeq&);
    ~InvocationHandler();
};

} // End of namespace OB

void OBDuplicate(OB::StubImplBase_ptr);
void OBRelease(OB::StubImplBase_ptr);

namespace CORBA
{

inline void
release(OB::StubImplBase_ptr p)
{
    if(p)
	p -> _OB_decRef();
}

inline Boolean
is_nil(OB::StubImplBase_ptr p)
{
    return p == 0;
}

} // End of namespace CORBA

#endif
